"""Router for the Job Script Template resource."""

from typing import cast

from buzz import require_condition, handle_errors
from fastapi import APIRouter, BackgroundTasks, Depends, File, HTTPException, Path, Query
from fastapi import Response as FastAPIResponse
from fastapi import UploadFile, status
from fastapi_pagination import Page
from loguru import logger
from pydantic import AnyUrl
import snick

from jobbergate_api.apps.constants import FileType
from jobbergate_api.apps.dependencies import SecureService, secure_services
from jobbergate_api.apps.garbage_collector import garbage_collector
from jobbergate_api.apps.job_script_templates.models import JobScriptTemplate
from jobbergate_api.apps.job_script_templates.tools import coerce_id_or_identifier
from jobbergate_api.apps.job_scripts.models import JobScriptFile
from jobbergate_api.apps.job_scripts.schemas import (
    JobScriptBaseView,
    JobScriptCloneRequest,
    JobScriptCreateRequest,
    JobScriptDetailedView,
    JobScriptFileDetailedView,
    JobScriptListView,
    JobScriptUpdateRequest,
    RenderFromTemplateRequest,
)
from jobbergate_api.apps.job_scripts.tools import inject_sbatch_params
from jobbergate_api.apps.permissions import Permissions, can_bypass_ownership_check
from jobbergate_api.apps.schemas import ListParams
from jobbergate_api.apps.services import ServiceError

router = APIRouter(prefix="/job-scripts", tags=["Job Scripts"])


@router.delete(
    "/clean-unused-entries",
    status_code=status.HTTP_202_ACCEPTED,
    description="Endpoint to automatically clean unused job scripts depending on a threshold",
    tags=["Garbage collector"],
)
def job_script_auto_clean_unused_entries(
    background_tasks: BackgroundTasks,
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_DELETE)
    ),
):
    """Automatically clean unused job scripts depending on a threshold."""
    logger.info("Starting automatically cleanup for unused job scripts")
    background_tasks.add_task(secure_services.crud.job_script.auto_clean_unused_job_scripts)
    return {"description": "Automatically cleanup started"}


@router.post(
    "",
    status_code=status.HTTP_201_CREATED,
    response_model=JobScriptBaseView,
    description="Endpoint for creating a stand alone job script. Use file upload to add files.",
)
async def job_script_create(
    create_request: JobScriptCreateRequest,
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_CREATE)
    ),
):
    """Create a stand alone job script."""
    logger.info(f"Creating a new job script with {create_request=}")

    if secure_services.identity_payload.email is None:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="The token payload does not contain an email",
        )

    return await secure_services.crud.job_script.create(
        owner_email=secure_services.identity_payload.email,
        **create_request.model_dump(exclude_unset=True),
    )


@router.post(
    "/clone/{id}",
    status_code=status.HTTP_201_CREATED,
    response_model=JobScriptDetailedView,
    description="Endpoint for cloning a job script to a new entry owned by the user",
)
async def job_script_clone(
    id: int = Path(),
    clone_request: JobScriptCloneRequest | None = None,
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_CREATE, ensure_email=True)
    ),
):
    """Clone a job script by its id."""
    logger.info(f"Cloning a job script from {id=} with {clone_request=}")

    if clone_request is None:
        clone_request = JobScriptCloneRequest()

    original_instance = await secure_services.crud.job_script.get(id, include_files=True)
    cloned_instance = await secure_services.crud.job_script.clone_instance(
        original_instance,
        owner_email=secure_services.identity_payload.email,
        **clone_request.model_dump(exclude_unset=True, exclude_none=True),
    )

    for file in original_instance.files:
        await secure_services.file.job_script.clone_instance(file, cloned_instance.id)

    # Get again to update the file data
    return await secure_services.crud.job_script.get(cloned_instance.id, include_files=True)


@router.post(
    "/render-from-template/{id_or_identifier}",
    status_code=status.HTTP_201_CREATED,
    response_model=JobScriptDetailedView,
    description="Endpoint for job script creation",
)
async def job_script_create_from_template(
    create_request: JobScriptCreateRequest,
    render_request: RenderFromTemplateRequest,
    id_or_identifier: str = Path(...),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_CREATE, ensure_email=True)
    ),
):
    """Create a new job script from a job script template."""
    typed_id_or_identifier = coerce_id_or_identifier(id_or_identifier)
    logger.info(f"Creating a new job script from {typed_id_or_identifier=} with {create_request=}")

    base_template = cast(
        JobScriptTemplate, await secure_services.crud.template.get(typed_id_or_identifier, include_files=True)
    )

    required_map = render_request.template_output_name_mapping

    required_keys = set(required_map.keys())
    provided_keys = set(f.filename for f in base_template.template_files)
    missing_keys = required_keys - provided_keys
    if len(missing_keys) > 0:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="The following required files are missing for job template {}: {}".format(
                typed_id_or_identifier, ", ".join(missing_keys)
            ),
        )

    mapped_template_files = {
        new_filename: file
        for file in base_template.template_files
        if (new_filename := required_map.get(file.filename, None))
    }

    entrypoint_files = [f for f in mapped_template_files.values() if f.file_type == FileType.ENTRYPOINT]
    if len(entrypoint_files) != 1:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"Exactly one entrypoint file must be specified, got {len(entrypoint_files)}",
        )

    job_script = await secure_services.crud.job_script.create(
        owner_email=secure_services.identity_payload.email,
        parent_template_id=base_template.id,
        **create_request.model_dump(exclude_unset=True),
    )

    for new_filename, template_file in mapped_template_files.items():
        file_content = await secure_services.file.template.render(
            template_file,
            render_request.param_dict,
        )

        if template_file.file_type == FileType.ENTRYPOINT and render_request.sbatch_params:
            with handle_errors(
                "Failed to inject sbatch params into the entrypoint file",
                raise_exc_class=ServiceError,
                raise_kwargs=dict(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY),
            ):
                file_content = inject_sbatch_params(file_content, render_request.sbatch_params)

        await secure_services.file.job_script.upsert(
            parent_id=job_script.id,
            filename=new_filename,
            upload_content=file_content,
            file_type=template_file.file_type,
        )

    return await secure_services.crud.job_script.get(job_script.id, include_files=True)


@router.get(
    "/{id}",
    description="Endpoint to return a job script by its id",
    response_model=JobScriptDetailedView,
)
async def job_script_get(
    id: int = Path(),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_READ, commit=False)
    ),
):
    """Get a job script by id."""
    logger.info(f"Getting job script {id=}")
    return await secure_services.crud.job_script.get(id, include_files=True)


@router.get(
    "",
    description="Endpoint to return a list of job scripts",
    response_model=Page[JobScriptListView],
)
async def job_script_get_list(
    list_params: ListParams = Depends(),
    from_job_script_template_id: int | None = Query(
        None,
        description="Filter job-scripts by the job-script-template-id they were created from.",
    ),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_READ, commit=False)
    ),
):
    """Get a list of job scripts."""
    logger.debug("Preparing to list job scripts")

    list_kwargs = list_params.model_dump(exclude_unset=True, exclude={"user_only"})
    list_kwargs["include_parent"] = True

    if from_job_script_template_id is not None:
        list_kwargs["parent_template_id"] = from_job_script_template_id
    if list_params.user_only:
        list_kwargs["owner_email"] = secure_services.identity_payload.email

    return await secure_services.crud.job_script.paginated_list(**list_kwargs)


@router.put(
    "/{id}",
    status_code=status.HTTP_200_OK,
    description="Endpoint to update a job script by id",
    response_model=JobScriptListView,
)
async def job_script_update(
    update_params: JobScriptUpdateRequest,
    id: int = Path(),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_UPDATE, ensure_email=True)
    ),
):
    """Update a job script template by id or identifier."""
    logger.info(f"Updating job script {id=} with {update_params=}")
    instance = await secure_services.crud.job_script.get(id, include_parent=True)
    if not can_bypass_ownership_check(secure_services.identity_payload.permissions):
        secure_services.crud.job_script.ensure_attribute(
            instance, owner_email=secure_services.identity_payload.email
        )
    return await secure_services.crud.job_script.update(id, **update_params.model_dump(exclude_unset=True))


@router.delete(
    "/{id}",
    status_code=status.HTTP_204_NO_CONTENT,
    description="Endpoint to delete a job script by id",
)
async def job_script_delete(
    id: int = Path(...),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_DELETE, ensure_email=True)
    ),
):
    """Delete a job script template by id or identifier."""
    logger.info(f"Deleting job script {id=}")
    instance = await secure_services.crud.job_script.get(id)
    if not can_bypass_ownership_check(secure_services.identity_payload.permissions):
        secure_services.crud.job_script.ensure_attribute(
            instance, owner_email=secure_services.identity_payload.email
        )
    await secure_services.crud.job_script.delete(id)
    return FastAPIResponse(status_code=status.HTTP_204_NO_CONTENT)


@router.get(
    "/{id}/upload/{file_name:path}",
    description="Endpoint to get a file from a job script",
)
async def job_script_get_file(
    id: int = Path(...),
    file_name: str = Path(...),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_READ, commit=False)
    ),
):
    """
    Get a job script file.

    Note:
        See https://fastapi.tiangolo.com/advanced/custom-response/#streamingresponse
    """
    job_script_file = await secure_services.file.job_script.get(id, file_name)
    return FastAPIResponse(
        content=await secure_services.file.job_script.get_file_content(job_script_file),
        media_type="text/plain",
        headers={"filename": job_script_file.filename},
    )


@router.put(
    "/{id}/upload-by-url/{file_type}",
    status_code=status.HTTP_200_OK,
    description=snick.unwrap(
        """
        Endpoint to upload a file to a job script from a provided URL.
        If a previous filename is provided, the file will be renamed from that.
        Upload file is optional in this scenario since the file content can be copied from previous file.
        """
    ),
    response_model=JobScriptFileDetailedView,
)
async def job_script_upload_file_by_url(
    id: int = Path(...),
    file_type: FileType = Path(...),
    filename: str | None = Query(None, max_length=255),
    file_url: AnyUrl = Query(..., description="URL of the file to upload"),
    previous_filename: str | None = Query(
        None,
        description="Previous name of the file in case a rename is needed",
        max_length=255,
    ),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_CREATE, ensure_email=True)
    ),
):
    """Upload a file to a job script from a URL."""
    if filename is None:
        url_path = file_url.path or ""
        (*_, filename) = url_path.split("/")
        require_condition(
            filename != "",
            f"Filename could not be extracted from the provided URL: {file_url}",
            raise_exc_class=HTTPException,
            exc_builder=lambda exc_class, msg: exc_class(status_code=status.HTTP_400_BAD_REQUEST, detail=msg),
        )

    # This is needed to make static type checkers happy. It shouldn't be able to happen
    assert filename is not None

    logger.debug(
        snick.unwrap(
            f"""
            Uploading file {filename=} from {file_url}
            to job script {id=}
            with {file_type=} and {previous_filename=}
            """
        )
    )

    job_script = await secure_services.crud.job_script.get(id)
    if not can_bypass_ownership_check(secure_services.identity_payload.permissions):
        secure_services.crud.job_script.ensure_attribute(
            job_script, owner_email=secure_services.identity_payload.email
        )

    return await secure_services.file.job_script.upsert(
        parent_id=job_script.id,
        filename=filename,
        upload_content=file_url,
        previous_filename=previous_filename,
        file_type=file_type,
    )


@router.put(
    "/{id}/upload/{file_type}",
    status_code=status.HTTP_200_OK,
    description=(
        "Endpoint to upload a file to a job script. "
        "If a previous filename is provided, the file will be renamed from that. "
        "Upload file is optional in this scenario since the file content can be copied from previous file."
    ),
    response_model=JobScriptFileDetailedView,
)
async def job_script_upload_file(
    id: int = Path(...),
    file_type: FileType = Path(...),
    filename: str | None = Query(None, max_length=255),
    upload_file: UploadFile | None = File(None, description="File to upload"),
    previous_filename: str | None = Query(
        None, description="Previous name of the file in case a rename is needed", max_length=255
    ),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_CREATE, ensure_email=True)
    ),
):
    """Upload a file to a job script."""
    # This is included for backwards compatibility with the previous implementation
    # where filename was recovered from the upload_file object
    filename = filename or getattr(upload_file, "filename")
    if not filename:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Filename must be provided either as a query parameter or as part of the file upload",
        )

    logger.debug(f"Uploading file {filename=} to job script {id=} with {file_type=} and {previous_filename=}")

    job_script = await secure_services.crud.job_script.get(id)
    if not can_bypass_ownership_check(secure_services.identity_payload.permissions):
        secure_services.crud.job_script.ensure_attribute(
            job_script, owner_email=secure_services.identity_payload.email
        )

    return await secure_services.file.job_script.upsert(
        parent_id=job_script.id,
        filename=filename,
        upload_content=upload_file,
        previous_filename=previous_filename,
        file_type=file_type,
    )


@router.delete(
    "/{id}/upload/{file_name}",
    status_code=status.HTTP_200_OK,
    description="Endpoint to delete a file from a job script",
)
async def job_script_delete_file(
    id: int = Path(...),
    file_name: str = Path(...),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_DELETE, ensure_email=True)
    ),
):
    """Delete a file from a job script template by id or identifier."""
    job_script = await secure_services.crud.job_script.get(id)
    if not can_bypass_ownership_check(secure_services.identity_payload.permissions):
        secure_services.crud.job_script.ensure_attribute(
            job_script, owner_email=secure_services.identity_payload.email
        )
    job_script_file = await secure_services.file.job_script.get(job_script.id, file_name)
    await secure_services.file.job_script.delete(job_script_file)


@router.delete(
    "/upload/garbage-collector",
    status_code=status.HTTP_202_ACCEPTED,
    description="Endpoint to delete all unused files from the job script file storage",
    tags=["Garbage collector"],
)
def job_script_garbage_collector(
    background_tasks: BackgroundTasks,
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_SCRIPTS_DELETE)
    ),
):
    """Delete all unused files from job scripts on the file storage."""
    logger.info("Starting garbage collection from jobbergate file storage")
    background_tasks.add_task(
        garbage_collector,
        secure_services.session,
        secure_services.bucket,
        [JobScriptFile],
        background_tasks,
    )
    return {"description": "Garbage collection started"}



"""Router for the Job Script Template resource."""

import snick
from buzz import require_condition
from fastapi import (
    APIRouter,
    BackgroundTasks,
    Body,
    Depends,
    File,
    HTTPException,
    Path,
    Query,
    UploadFile,
    status,
)
from fastapi import Response as FastAPIResponse
from fastapi_pagination import Page
from loguru import logger
from pydantic import AnyUrl
from sqlalchemy.exc import IntegrityError

from jobbergate_api.apps.constants import FileType
from jobbergate_api.apps.dependencies import SecureService, secure_services
from jobbergate_api.apps.garbage_collector import garbage_collector
from jobbergate_api.apps.job_script_templates.constants import WORKFLOW_FILE_NAME
from jobbergate_api.apps.job_script_templates.models import JobScriptTemplateFile, WorkflowFile
from jobbergate_api.apps.job_script_templates.schemas import (
    JobTemplateBaseDetailedView,
    JobTemplateCloneRequest,
    JobTemplateCreateRequest,
    JobTemplateDetailedView,
    JobTemplateListView,
    JobTemplateUpdateRequest,
    RunTimeConfig,
    TemplateFileDetailedView,
    WorkflowFileDetailedView,
)
from jobbergate_api.apps.job_script_templates.tools import coerce_id_or_identifier
from jobbergate_api.apps.permissions import Permissions, can_bypass_ownership_check
from jobbergate_api.apps.schemas import ListParams
from jobbergate_api.apps.services import ServiceError

router = APIRouter(prefix="/job-script-templates", tags=["Job Script Templates"])


@router.post(
    "",
    status_code=status.HTTP_201_CREATED,
    response_model=JobTemplateBaseDetailedView,
    description="Endpoint for job script template creation",
)
async def job_script_template_create(
    create_request: JobTemplateCreateRequest,
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_TEMPLATES_CREATE, ensure_email=True)
    ),
):
    """Create a new job script template."""
    logger.info(f"Creating a new job script template with {create_request=}")

    try:
        return await secure_services.crud.template.create(
            owner_email=secure_services.identity_payload.email,
            **create_request.model_dump(exclude_unset=True),
        )
    except IntegrityError:
        message = f"Job script template with the identifier={create_request.identifier} already exists"
        logger.error(message)
        raise HTTPException(status_code=status.HTTP_409_CONFLICT, detail=message)


@router.get(
    "/{id_or_identifier}",
    description="Endpoint to return a job script template by its id or identifier",
    response_model=JobTemplateDetailedView,
)
async def job_script_template_get(
    id_or_identifier: str = Path(),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_TEMPLATES_READ, commit=False)
    ),
):
    """Get a job script template by id or identifier."""
    typed_id_or_identifier: int | str = coerce_id_or_identifier(id_or_identifier)
    logger.info(f"Getting job script template with {typed_id_or_identifier=}")
    return await secure_services.crud.template.get(typed_id_or_identifier, include_files=True)


@router.post(
    "/clone/{id_or_identifier}",
    status_code=status.HTTP_201_CREATED,
    response_model=JobTemplateDetailedView,
    description="Endpoint for cloning a job script template to a new entry owned by the user",
)
async def job_script_template_clone(
    id_or_identifier: str = Path(),
    clone_request: JobTemplateCloneRequest | None = None,
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_TEMPLATES_CREATE, ensure_email=True)
    ),
):
    """Clone a job script template by id or identifier."""
    typed_id_or_identifier: int | str = coerce_id_or_identifier(id_or_identifier)
    logger.info(f"Cloning a job script template from {typed_id_or_identifier=} with {clone_request=}")

    if clone_request is None:
        clone_request = JobTemplateCloneRequest()

    original_instance = await secure_services.crud.template.get(typed_id_or_identifier, include_files=True)

    # Identifier is specifically set to None to avoid conflicts with the original instance
    new_data = {"identifier": None, **clone_request.model_dump(exclude_unset=True)}
    try:
        cloned_instance = await secure_services.crud.template.clone_instance(
            original_instance,
            owner_email=secure_services.identity_payload.email,
            **new_data,
        )
    except IntegrityError:
        message = "Job script template with the identifier={} already exists".format(
            clone_request.identifier or original_instance.identifier
        )
        logger.error(message)
        raise HTTPException(status_code=status.HTTP_409_CONFLICT, detail=message)

    for file in original_instance.template_files:
        await secure_services.file.template.clone_instance(file, cloned_instance.id)
    for file in original_instance.workflow_files:
        await secure_services.file.workflow.clone_instance(file, cloned_instance.id)

    # Get again to update the file data
    return await secure_services.crud.template.get(cloned_instance.id, include_files=True)


@router.get(
    "",
    description="Endpoint to return a list of job script templates",
    response_model=Page[JobTemplateListView],
)
async def job_script_template_get_list(
    list_params: ListParams = Depends(),
    include_null_identifier: bool = Query(False),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_TEMPLATES_READ, commit=False)
    ),
):
    """Get a list of job script templates."""
    logger.debug("Preparing to list job script templates")

    list_kwargs = list_params.model_dump(exclude_unset=True, exclude={"user_only", "include_parent"})

    if list_params.user_only:
        list_kwargs["owner_email"] = secure_services.identity_payload.email

    return await secure_services.crud.template.paginated_list(
        **list_kwargs,
        include_null_identifier=include_null_identifier,
    )


@router.put(
    "/{id_or_identifier}",
    status_code=status.HTTP_200_OK,
    description="Endpoint to update a job script template by id or identifier",
    response_model=JobTemplateBaseDetailedView,
)
async def job_script_template_update(
    update_request: JobTemplateUpdateRequest,
    id_or_identifier: str = Path(),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_TEMPLATES_UPDATE, ensure_email=True)
    ),
):
    """Update a job script template by id or identifier."""
    typed_id_or_identifier: int | str = coerce_id_or_identifier(id_or_identifier)
    logger.info(f"Updating job script template {typed_id_or_identifier=} with {update_request=}")
    instance = await secure_services.crud.template.get(typed_id_or_identifier)
    if not can_bypass_ownership_check(secure_services.identity_payload.permissions):
        secure_services.crud.template.ensure_attribute(
            instance, owner_email=secure_services.identity_payload.email
        )
    return await secure_services.crud.template.update(
        typed_id_or_identifier, **update_request.model_dump(exclude_unset=True)
    )


@router.delete(
    "/{id_or_identifier}",
    status_code=status.HTTP_204_NO_CONTENT,
    description="Endpoint to delete a job script template by id or identifier",
)
async def job_script_template_delete(
    id_or_identifier: str = Path(),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_TEMPLATES_DELETE, ensure_email=True)
    ),
):
    """Delete a job script template by id or identifier."""
    typed_id_or_identifier: int | str = coerce_id_or_identifier(id_or_identifier)
    logger.info(f"Deleting job script template with {typed_id_or_identifier=}")
    isinstance = await secure_services.crud.template.get(typed_id_or_identifier)
    if not can_bypass_ownership_check(secure_services.identity_payload.permissions):
        secure_services.crud.template.ensure_attribute(
            isinstance, owner_email=secure_services.identity_payload.email
        )
    await secure_services.crud.template.delete(typed_id_or_identifier)
    return FastAPIResponse(status_code=status.HTTP_204_NO_CONTENT)


@router.get(
    "/{id_or_identifier}/upload/template/{file_name:path}",
    description="Endpoint to get a file from a job script template by id or identifier",
)
async def job_script_template_get_file(
    id_or_identifier: str = Path(),
    file_name: str = Path(),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_TEMPLATES_READ, commit=False)
    ),
):
    """
    Get a job script template file by id or identifier.

    Note:
        See https://fastapi.tiangolo.com/advanced/custom-response/#streamingresponse
    """
    typed_id_or_identifier: int | str = coerce_id_or_identifier(id_or_identifier)
    logger.debug(f"Getting template file {file_name=} from job script template {typed_id_or_identifier=}")
    job_script_template = await secure_services.crud.template.get(typed_id_or_identifier)
    job_script_template_file = await secure_services.file.template.get(job_script_template.id, file_name)
    return FastAPIResponse(
        content=await secure_services.file.template.get_file_content(job_script_template_file),
        media_type="text/plain",
        headers={"filename": job_script_template_file.filename},
    )


async def _upsert_template_file(
    id_or_identifier: str,
    file_type: FileType,
    filename: str,
    upload_content: str | bytes | AnyUrl | UploadFile | None,
    previous_filename: str | None,
    secure_services: SecureService,
):
    """
    Provide an auxillary function to be used for uploading from file object or URL.
    """
    typed_id_or_identifier: int | str = coerce_id_or_identifier(id_or_identifier)
    job_script_template = await secure_services.crud.template.get(typed_id_or_identifier)
    if not can_bypass_ownership_check(secure_services.identity_payload.permissions):
        secure_services.crud.template.ensure_attribute(
            job_script_template, owner_email=secure_services.identity_payload.email
        )

    return await secure_services.file.template.upsert(
        parent_id=job_script_template.id,
        filename=filename,
        upload_content=upload_content,
        previous_filename=previous_filename,
        file_type=file_type,
    )


@router.put(
    "/{id_or_identifier}/upload/template/{file_type}",
    status_code=status.HTTP_200_OK,
    description=(
        "Endpoint to upload a file to a job script template by id or identifier. "
        "If a previous filename is provided, the file will be renamed from that. "
        "Upload file is optional in this scenario since the file content can be copied from previous file."
    ),
    response_model=TemplateFileDetailedView,
)
async def job_script_template_upload_file(
    id_or_identifier: str = Path(),
    file_type: FileType = Path(),
    filename: str | None = Query(None, max_length=255),
    upload_file: UploadFile | None = File(None, description="File to upload"),
    previous_filename: str | None = Query(
        None, description="Previous name of the file in case a rename is needed", max_length=255
    ),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_TEMPLATES_CREATE, ensure_email=True)
    ),
):
    """Upload a file to a job script template by id or identifier."""
    # This is included for backwards compatibility with the previous implementation
    # where filename was recovered from the upload_file object
    filename = filename or getattr(upload_file, "filename")
    if not filename:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Filename must be provided either as a query parameter or as part of the file upload",
        )

    logger.debug(
        f"Uploading {filename=} to job template {id_or_identifier=}; {file_type=}; {previous_filename=}"
    )
    return await _upsert_template_file(
        id_or_identifier,
        file_type,
        filename,
        upload_file,
        previous_filename,
        secure_services,
    )


@router.put(
    "/{id_or_identifier}/upload-by-url/template/{file_type}",
    status_code=status.HTTP_200_OK,
    description=(
        "Endpoint to upload a file to a job script template by id or identifier using a file URL. "
        "If a previous filename is provided, the file will be renamed from that. "
    ),
    response_model=TemplateFileDetailedView,
)
async def job_script_template_upload_file_by_url(
    id_or_identifier: str = Path(),
    file_type: FileType = Path(),
    filename: str | None = Query(None, max_length=255),
    file_url: AnyUrl = Query(..., description="URL of the file to upload"),
    previous_filename: str | None = Query(
        None, description="Previous name of the file in case a rename is needed", max_length=255
    ),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_TEMPLATES_CREATE, ensure_email=True)
    ),
):
    """Upload a file to a job script template by id or identifier using file URL."""

    if filename is None:
        url_path = file_url.path or ""
        (*_, filename) = url_path.split("/")
        require_condition(
            filename != "",
            f"Filename could not be extracted from the provided URL: {file_url}",
            raise_exc_class=HTTPException,
            exc_builder=lambda exc_class, msg: exc_class(status_code=status.HTTP_400_BAD_REQUEST, detail=msg),
        )

    # This is needed to make static type checkers happy. It shouldn't be able to happen
    assert filename is not None

    logger.debug(
        snick.unwrap(
            f"""
            Uploading file {filename=} from {file_url}
            to job template {id_or_identifier=};
            {file_type=}; {previous_filename=}
            """
        )
    )
    return await _upsert_template_file(
        id_or_identifier,
        file_type,
        filename,
        file_url,
        previous_filename,
        secure_services,
    )


@router.delete(
    "/{id_or_identifier}/upload/template/{file_name:path}",
    status_code=status.HTTP_200_OK,
    description="Endpoint to delete a file to a job script template by id or identifier",
)
async def job_script_template_delete_file(
    id_or_identifier: str = Path(),
    file_name: str = Path(),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_TEMPLATES_DELETE, ensure_email=True)
    ),
):
    """Delete a file from a job script template by id or identifier."""
    typed_id_or_identifier: int | str = coerce_id_or_identifier(id_or_identifier)
    job_script_template = await secure_services.crud.template.get(typed_id_or_identifier)
    if not can_bypass_ownership_check(secure_services.identity_payload.permissions):
        secure_services.crud.template.ensure_attribute(
            job_script_template, owner_email=secure_services.identity_payload.email
        )
    job_script_template_file = await secure_services.file.template.get(job_script_template.id, file_name)
    await secure_services.file.template.delete(job_script_template_file)


@router.get(
    "/{id_or_identifier}/upload/workflow",
    description="Endpoint to get a workflow file from a job script template by id or identifier",
)
async def job_script_workflow_get_file(
    id_or_identifier: str = Path(),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_TEMPLATES_READ, commit=False)
    ),
):
    """
    Get a workflow file by id or identifier.

    Note:
        See https://fastapi.tiangolo.com/advanced/custom-response/#streamingresponse
    """
    typed_id_or_identifier: int | str = coerce_id_or_identifier(id_or_identifier)
    logger.debug(f"Getting workflow file from job script template {typed_id_or_identifier=}")
    job_script_template = await secure_services.crud.template.get(typed_id_or_identifier)
    workflow_file = await secure_services.file.workflow.get(job_script_template.id, WORKFLOW_FILE_NAME)
    return FastAPIResponse(
        content=await secure_services.file.workflow.get_file_content(workflow_file),
        media_type="text/plain",
        headers={"filename": WORKFLOW_FILE_NAME},
    )


async def _upsert_workflow_file(
    id_or_identifier: str,
    runtime_config: RunTimeConfig | None,
    upload_content: str | bytes | AnyUrl | UploadFile,
    secure_services: SecureService,
):
    """
    Provide an auxillary function to be used for uploading from file object or URL.
    """
    typed_id_or_identifier: int | str = coerce_id_or_identifier(id_or_identifier)
    logger.debug(
        f"Uploading workflow file to job script template {typed_id_or_identifier=}: {runtime_config}"
    )
    job_script_template = await secure_services.crud.template.get(typed_id_or_identifier)
    if not can_bypass_ownership_check(secure_services.identity_payload.permissions):
        secure_services.crud.template.ensure_attribute(
            job_script_template, owner_email=secure_services.identity_payload.email
        )
    upsert_kwargs = dict(
        parent_id=job_script_template.id, filename=WORKFLOW_FILE_NAME, upload_content=upload_content
    )

    try:
        await secure_services.file.workflow.get(job_script_template.id, WORKFLOW_FILE_NAME)
        file_exist = True
    except ServiceError:
        file_exist = False

    if runtime_config is None and file_exist is False:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Runtime configuration is required when the workflow file does not exist",
        )
    elif runtime_config is not None:
        upsert_kwargs["runtime_config"] = runtime_config.model_dump()

    return await secure_services.file.workflow.upsert(**upsert_kwargs)


@router.put(
    "/{id_or_identifier}/upload/workflow",
    status_code=status.HTTP_200_OK,
    description="Endpoint to upload a file to a job script template by id or identifier",
    response_model=WorkflowFileDetailedView,
)
async def job_script_workflow_upload_file(
    id_or_identifier: str = Path(),
    runtime_config: RunTimeConfig | None = Body(
        None, description="Runtime configuration is optional when the workflow file already exists"
    ),
    upload_file: UploadFile = File(..., description="File to upload"),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_TEMPLATES_CREATE, ensure_email=True)
    ),
):
    """Upload a file to a job script workflow by id or identifier."""
    return await _upsert_workflow_file(id_or_identifier, runtime_config, upload_file, secure_services)


@router.put(
    "/{id_or_identifier}/upload-by-url/workflow",
    status_code=status.HTTP_200_OK,
    description="Endpoint to upload a file to a job script template by id or identifier from a provided URL.",
    response_model=WorkflowFileDetailedView,
)
async def job_script_upload_file_by_url(
    id_or_identifier: str = Path(),
    runtime_config: RunTimeConfig | None = Body(
        None, description="Runtime configuration is optional when the workflow file already exists"
    ),
    file_url: AnyUrl = Query(..., description="URL of the file to upload"),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_TEMPLATES_CREATE, ensure_email=True)
    ),
):
    """Upload a file to a job script workflow by id or identifier from a URL."""
    return await _upsert_workflow_file(id_or_identifier, runtime_config, file_url, secure_services)


@router.delete(
    "/{id_or_identifier}/upload/workflow",
    status_code=status.HTTP_200_OK,
    description="Endpoint to delete a workflow file from a job script template by id or identifier",
)
async def job_script_workflow_delete_file(
    id_or_identifier: str = Path(),
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_TEMPLATES_DELETE, ensure_email=True)
    ),
):
    """Delete a workflow file from a job script template by id or identifier."""
    typed_id_or_identifier: int | str = coerce_id_or_identifier(id_or_identifier)
    logger.debug(f"Deleting workflow file from job script template {typed_id_or_identifier=}")
    job_script_template = await secure_services.crud.template.get(typed_id_or_identifier)
    if not can_bypass_ownership_check(secure_services.identity_payload.permissions):
        secure_services.crud.template.ensure_attribute(
            job_script_template, owner_email=secure_services.identity_payload.email
        )
    workflow_file = await secure_services.file.workflow.get(job_script_template.id, WORKFLOW_FILE_NAME)
    await secure_services.file.workflow.delete(workflow_file)


@router.delete(
    "/upload/garbage-collector",
    status_code=status.HTTP_202_ACCEPTED,
    description="Endpoint to delete all unused files from the job script template file storage",
    tags=["Garbage collector"],
)
async def job_script_template_garbage_collector(
    background_tasks: BackgroundTasks,
    secure_services: SecureService = Depends(
        secure_services(Permissions.ADMIN, Permissions.JOB_TEMPLATES_DELETE)
    ),
):
    """Delete all unused files from jobbergate templates on the file storage."""
    logger.info("Starting garbage collection from jobbergate file storage")
    background_tasks.add_task(
        garbage_collector,
        secure_services.session,
        secure_services.bucket,
        [JobScriptTemplateFile, WorkflowFile],
        background_tasks,
    )
    return {"description": "Garbage collection started"}



"""Schema definitions for the cluster app."""

from pydantic import BaseModel, ConfigDict

from jobbergate_api.apps.schemas import PydanticDateTime
from jobbergate_api.meta_mapper import MetaField, MetaMapper

cluster_status_meta_mapper = MetaMapper(
    client_id=MetaField(
        description="The client_id of the cluster where this agent is reporting status from",
        example="mega-cluster-1",
    ),
    created_at=MetaField(
        description="The timestamp for when this entry was created",
        example="2023-08-18T13:55:37.172285",
    ),
    updated_at=MetaField(
        description="The timestamp for when this entry was last updated",
        example="2023-08-18T13:55:37.172285",
    ),
    last_reported=MetaField(
        description="The timestamp for when the agent on the cluster last reported its status",
        example="2023-08-18T13:55:37.172285",
    ),
    interval=MetaField(
        description="The expected interval in seconds between pings from the agent",
        example=60,
    ),
    is_healthy=MetaField(
        description="A boolean indicating if the cluster is healthy based on the last_reported time",
        example=True,
    ),
)


class ClusterStatusView(BaseModel):
    """
    Describes the status of a cluster.
    """

    client_id: str
    created_at: PydanticDateTime
    updated_at: PydanticDateTime
    last_reported: PydanticDateTime
    interval: int
    is_healthy: bool

    model_config = ConfigDict(
        from_attributes=True,
        json_schema_extra=cluster_status_meta_mapper,
    )



"""
Validation methods for the uploaded files.
"""

from ast import parse as ast_parse
from functools import wraps
from pathlib import PurePath
from typing import BinaryIO, Callable, Union

from buzz import require_condition
from jinja2 import Environment
from loguru import logger
from yaml import safe_load as yaml_safe_load


def get_suffix(filename: str) -> str:
    """
    Get the suffix (file extension) from a given filename.
    """
    return PurePath(filename).suffix


def check_uploaded_file_syntax(file_obj: BinaryIO, filename: str) -> bool:
    """
    Check the syntax of a given file.
    """
    logger.debug(f"Validating source code on {filename=}")
    suffix = get_suffix(filename)
    if suffix in syntax_validation_dispatch:
        validator = syntax_validation_dispatch[suffix]
        result = validator(file_obj.read())
        file_obj.seek(0)
        return result
    logger.debug(f"Skipping because {suffix} has no syntax validation")
    return True


SyntaxValidationEquation = Callable[[Union[str, bytes]], bool]
"""Type alias describing the function signature used to validate file syntax."""

syntax_validation_dispatch: dict[str, SyntaxValidationEquation] = {}
"""Dictionary mapping file extensions to the function used to validate their syntax."""


def register_syntax_validator(*file_extensions: str):
    """
    Use this decorator to register file syntax validation functions.

    It creates a new entry on ``validation_dispatch``, mapping the equation to
    the file extensions that are provided as arguments.

    Raise ValueError if the provided file extensions do not start with a dot.
    """

    def decorator(validator):
        for extension in file_extensions:
            require_condition(
                extension.startswith("."),
                f"File extensions should start with a dot, got the value: {extension}",
                raise_exc_class=ValueError,
            )
            syntax_validation_dispatch[extension] = validator

        @wraps(validator)
        def wrapper(source_code):
            return validator(source_code)

        return wrapper

    return decorator


@register_syntax_validator(".py")
def is_valid_python_file(source_code: Union[str, bytes]) -> bool:
    """
    Check if a given Python source code is valid by parsing it into an AST node.

    Args:
        source_code: Python source code.

    Returns:
        Boolean indicating if the source code is valid or not.
    """
    try:
        ast_parse(source_code)
    except Exception:
        return False
    return True


@register_syntax_validator(".yaml")
def is_valid_yaml_file(yaml_file: Union[str, bytes]) -> bool:
    """
    Check if a given YAML file is valid by parsing it with yaml.safe_load.

    Args:
        yaml_file: YAML file.

    Returns:
        Boolean indicating if the file is valid or not.
    """
    try:
        yaml_safe_load(yaml_file)
    except Exception:
        return False
    return True


@register_syntax_validator(".j2", ".jinja2")
def is_valid_jinja2_template(template: Union[str, bytes]) -> bool:
    """
    Check if a given jinja2 template is valid by creating a Template object and trying to render it.

    Args:
        template: Jinja2 template.

    Returns:
        Boolean indicating if the template is valid or not.
    """
    if isinstance(template, bytes):
        _template = template.decode("utf-8")
    else:
        _template = template

    try:
        Environment().parse(_template)
    except Exception:
        return False
    return True
